module Quandl::Cassandra::Models::Data::Search
  
  extend ActiveSupport::Concern
  
  included do
    
    include ScopeComposer::Model
  
    def self.scope_names
      scope.scope_names
    end
  
    has_scope_composer
  
    delegate :where, to: :scope
  
    scope :dataset, ->(d){
      id(d.id)
      column_ids(d.column_ids)
    }
  
    scope :row, :id, :limit, :offset, :column, :accuracy, :frequency, :count, :delete
    
    scope :column_frequencies, ->(*freqs){ where( column_frequencies: Array(freqs).flatten ) }
    scope :column_ids, ->(*ids){ cids = Array(ids).flatten.compact; where( column_ids: cids ) if cids.present? }

    scope :collapse,  ->(v){ where( collapse:   v.to_sym ) if Quandl::Operation::Collapse.valid_collapse?(v) }
    scope :transform, ->(v){ where( transform:  v.to_sym ) if Quandl::Operation::Transform.valid_transformation?(v) }
  
    scope :order, ->(v){
      order = ( v.to_sym == :asc ) ? :asc : :desc
      where( order: order )
    }
  
    scope :trim_start,  ->(date){ date = parse_date(date); where( trim_start: date ) if date }
    scope :trim_end,    ->(date){ date = parse_date(date); where( trim_end:   date ) if date }
  
    scope_helper :find,       ->(id){ id(id).to_table }
    scope_helper :to_table,   ->{ all }
  
    scope_helper :parse_date, ->( value ){ 
      begin
        date = Date.jd(value.to_i) if value.kind_of?(String) && value.numeric?
        date = Date.jd(value) if value.is_a?(Integer)
        date = Date.parse(value) if value.is_a?(String) && value =~ /^[0-9]{4}\-[0-9]{2}\-[0-9]{2}$/
        date = value if value.is_a?(Date)
        date = value.to_date if value.respond_to?(:to_date)
        date.jd
      rescue
        nil
      end
    }
    
    scope.class_eval do
      
      delegate :to_a, :==, :inspect, :first, :flatten, :[], :collect, :<=>, :each, :each_with_index, :to_date, :to_h, to: :all, allow_nil: true
      
      def delete_all(*args)
        if dataset_column_ids.present?
          result = Quandl::Cassandra::Models::Column.where( id: dataset_column_ids ).delete_all
          return result.nil? ? true : result
        end
        false
      end
  
      def delete_rows_by_time(*times)
        times = Array(times).flatten
        # parse dates
        times = times.collect{|time| parse_date(time) }.compact
        # dataset_column_ids are required
        return false if times.blank?
        return false unless dataset_column_ids
        statement_values = []
        # for each column_id
        dataset_column_ids.each do |column_id|
          # for each time
          times.each do |time|
            # delete the requested column_id_time
            statement_values << [column_id, time]
          end
        end
        # execute deletion of rquested rows
        Quandl::Cassandra::Batch.execute(statement_values) do |id, time|
          "DELETE FROM columns WHERE id = #{id} AND type = 'source' AND time = #{time}"
        end
        # delete collapse data and regenerate
        regenerate_collapses_from_source
      end
    
      def count(*args)
        attributes[:count] = true
        result = dataset? ? fetch.to_i : 0
        attributes[:count] = false
        result
      end
      
      def all
        @all ||= fetch
      end
      
      def dataset?
        self.id.present? || dataset_column_ids.present?
      end
      
      def fetched?
        @all.present?
      end
      
      def scoped
        s = self.class.new
        s.id( self.id ) if id.present?
        s.column_ids( dataset_column_ids ) if dataset_column_ids.present?
        s.column_frequencies(attributes[:column_frequencies]) if attributes[:column_frequencies].present?
        s
      end
      
      def dataset_column_ids
        attributes[:column_ids]
      end
      
      protected
      
      def fetch
        # without an id or columns there's nothing to be read
        return Quandl::Cassandra::Models::Data.new unless dataset?
        # otherwise read the data
        Quandl::Cassandra::Models::Column.read( attributes.merge(scope_attributes) )
      end
      
      private
      
      def regenerate_collapses_from_source
        # delete collapse data
        delete_collapses
        # get existing data
        data = Quandl::Cassandra::Models::Column.read( column_ids: dataset_column_ids )
        # rewrite data
        Quandl::Cassandra::Models::Column.write( id: id, data: data, column_ids: dataset_column_ids )
      end
    
      def delete_collapses
        statement_values = []
        # for each column
        dataset_column_ids.each do |column_id|
          # for each collapse
          Quandl::Operation::Collapse.collapses_greater_than(:daily).each do |collapse|
            statement_values << [ column_id, collapse ]
          end
        end
        # execute deletion
        Quandl::Cassandra::Batch.execute(statement_values) do |id, type|
          "DELETE FROM columns WHERE id = #{id} AND type = '#{type}'"
        end
      end
      
    end
    
  end
end